!------------------------------------------------------------------------------
!                             Copyright,  2009
!                                Greg Howes
!------------------------------------------------------------------------------
!          HYDRO: SAMPLE PARALLEL CODE FOR HYDRODYNAMIC SIMULATION
!------------------------------------------------------------------------------
! NOTES:
program hydro
  use hydro_param, only: init_param,runname,nsteps,nsave,dt,ny
  use hydro_mpi_var, only: nproc,iproc,proc0,ierror
  use hydro_grid, only: setup_grid,finalize_grid,lnx0,lnx1
  use hydro_ic, only: setup_ic,gamma,cs0
  use hydro_dt, only: init_ab3_constants,adv_ab3,init_twosteps
  use hydro_var, only: qq,dqdt,dqdt1,dqdt2,rx,ry,imax,tt
  use hydro_hd, only: dqdt_hd
  use hydro_bcs, only: pass_bcs
  use hydro_diag, only: save_output
  implicit none
  include 'mpif.h'      !Include MPI library variables
  include 'pat_apif.h'  !CrayPAT
  integer :: istat      !CrayPAT
  integer :: it         !Step number
  integer :: ix,iy      !Counting indices for grid
  logical :: finish=.false.     !Flag to quit program

  !Initialize MPI message passing---------------------------------------------
  call mpi_init (ierror)
  call mpi_comm_size (mpi_comm_world, nproc, ierror)
  call mpi_comm_rank (mpi_comm_world, iproc, ierror)
  !Set logical proc0=true if iproc=0
  proc0= (iproc == 0)

  call PAT_tracing_state(PAT_STATE_ON,istat)

  if (proc0) write(*,'(a)')'Starting HYDRO===================================' 
  
  !Check to be sure nproc is even, otherwise shutdown
  if (mod(nproc,2) .ne. 0) then
     if (proc0) then
        write(*,'(a,i6)')'ERROR: Number of processors must be even: nproc= ',nproc
        write(*,'(a)')   'Finishing HYDRO==================================' 
     endif
     call mpi_finalize (ierror)
     stop
  endif

  !Read parameters------------------------------------------------------------
  call init_param
  if (proc0) write(*,'(a,a)') 'Input File name: ',runname

  !TEST
  if (.false.) then
     write(*,'(a,i6)') 'Hello, I am processor number ',iproc
  endif

  !Set up grid----------------------------------------------------------------
  call setup_grid

  !Set up initial conditions--------------------------------------------------
  call setup_ic
  !Output initial conditions
  call save_output(0)

  !Initialize Timestepping for AB3 time advance (step=0 and step =1)----------
  if (proc0) write(*,'(a)')'>>> BEGIN: Initialization of timesteps-----------' 
  call init_ab3_constants
  call init_twosteps
  if (proc0) write(*,'(a)')'<<< COMPLETE: Initialization of timesteps--------' 
  
!PAT Tracing
!  call PAT_tracing_state(PAT_STATE_ON,istat)
!  call PAT_region_begin(100,'Main advance loop',istat)


  if (.true.) then
  !MAIN PROGRAM TIMESTEP LOOP=================================================
  do it=3,nsteps
     !Pass Boundary Condition information among processors--------------------
     call pass_bcs

     !Calculate Time derivatives of fluid variables---------------------------
     !Move old timesteps back one
     dqdt2=dqdt1
     dqdt1=dqdt
     !Loop over all local points to calculate current derivatives dqdt
     do ix=lnx0,lnx1 !Loop over all local ix's
        do iy=0,ny-1
           dqdt(1:imax,ix,iy)=dqdt_hd(qq(1:imax,ix-1:ix+1,iy-1:iy+1), &
                rx(ix-1:ix+1),ry(iy-1:iy+1),gamma)
        enddo
     enddo

     !Time step fluid variables forward using AB3-----------------------------
     do ix=lnx0,lnx1 !Loop over all local ix's
        do iy=0,ny-1
           qq(1:imax,ix,iy)=adv_ab3(qq(1:imax,ix,iy),dt,dqdt(1:imax,ix,iy), &
                dqdt1(1:imax,ix,iy),dqdt2(1:imax,ix,iy))
        enddo
     enddo
     
     !Advance time
     tt=tt+dt
     if (proc0) write(*,'(a,i6,a,es12.4)')'Timestep ',it,': t=',tt
     
     !Check if we need to save------------------------------------------------
     if (mod(it,nsave) .eq. 0) then
        call save_output(it)
     endif

     !Check if we have signaled to exit the program---------------------------
     call checkstop(finish)

     if (finish .or. it .eq. nsteps) then
!        call final_diagnostics
        exit !Jump out of main timestep loop
     endif
     
  enddo !END MAIN PROGRAM LOOP================================================
  endif
  
!  call PAT_region_end(100,istat)
  call PAT_flush_buffer(istat)
  call PAT_tracing_state(PAT_STATE_OFF,istat)

!  call mpi_barrier(mpi_comm_world)

  if (proc0) write(*,'(a)')'Finishing HYDRO==================================' 
  call finalize_grid

  !Finalize MPI message passing
  call mpi_finalize (ierror)

contains
!-----------------------------------------------------------------------------
!                           Greg Howes, 2009
!------------------------------------------------------------------------------
! Check if the file runname.stop has been created. If so, finish1=T.
  subroutine checkstop(finish1)
    implicit none
    logical, intent (in out) :: finish1
    character (len=300) :: filename
    
    logical :: exit_local
    
    ! If .stop file has appeared, set finish1 flag
    filename=trim(runname)//".stop"
    
    if (proc0) then
       inquire(file=filename,exist=exit_local)
       finish1 = finish1 .or. exit_local
    end if

    call mpi_bcast (finish1, 1, MPI_LOGICAL, 0, mpi_comm_world, ierror)

  end subroutine checkstop
!------------------------------------------------------------------------------
end program hydro
